@@ -10,7 +10,7 @@
    to maintain the source code under the licensing terms described
    herein and below.
 
-   Copyright © 2021 Logan Rathbone <poprocks@gmail.com>
+   Copyright © 2021-2023 Logan Rathbone <poprocks@gmail.com>
 
    GHex is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -38,6 +38,8 @@
 
 #include <config.h>
 
+G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+
 /* Don't move these from the source file as they are not part of the public
  * header.
  */
@@ -545,6 +547,20 @@ geometry_action (GtkWidget *widget,
 	gtk_popover_popup (GTK_POPOVER(self->geometry_popover));
 }
 
+static void
+move_to_buffer_ends_action (GtkWidget *widget,
+		const char *action_name,
+		GVariant *parameter)
+{
+	HexWidget *self = HEX_WIDGET (widget);
+	gboolean beginning, extend_selection;
+
+	g_variant_get (parameter, "(bb)", &beginning, &extend_selection);
+
+	self->selecting = extend_selection;
+	hex_widget_set_cursor (self, beginning ? 0 : HEX_BUFFER_PAYLOAD (self->document));
+}
+
 /*
  * ?_to_pointer translates mouse coordinates in hex/ascii view
  * to cursor coordinates.
@@ -1544,6 +1560,57 @@ update_primary_selection (HexWidget *self)
 	}
 }
 
+inline static size_t
+clamp_rep_len (HexWidget *self, size_t start_pos, size_t rep_len)
+{
+	size_t payload_size = HEX_BUFFER_PAYLOAD (self->document);
+
+	if (rep_len > payload_size - start_pos)
+		rep_len = payload_size - start_pos;
+
+	return rep_len;
+}
+
+static void
+paste_set_data (HexWidget *self, char *data, size_t data_len)
+{
+	size_t start_pos = self->cursor_pos;
+	size_t len = data_len;
+	size_t rep_len = 0;
+
+	if (self->selection.start != self->selection.end) {
+		size_t selection_len, end_pos;
+
+		start_pos = MIN(self->selection.start, self->selection.end);
+		end_pos = MAX(self->selection.start, self->selection.end);
+		selection_len = end_pos - start_pos + 1;
+
+		if (self->insert) {
+			rep_len = selection_len;
+		} else {
+			len = rep_len = MIN(selection_len, len);
+		}
+	}
+
+	if (! self->insert)
+		len = rep_len = clamp_rep_len (self, start_pos, len);
+
+	hex_document_set_data (self->document,
+			start_pos,
+			len,
+			/* rep_len is either:
+			 * 0 (insert w/o replacing),
+			 * len (delete and paste the same number of bytes),
+			 * or a different number (delete and paste an arbitrary number
+			 * of bytes)
+			 */
+			rep_len,
+			data,
+			TRUE);
+
+	hex_widget_set_cursor (self, start_pos + len);
+}
+
 static void
 plaintext_paste_received_cb (GObject *source_object,
 		GAsyncResult *result,
@@ -1562,18 +1629,13 @@ plaintext_paste_received_cb (GObject *source_object,
 	/* Get the resulting text of the read operation */
 	text = gdk_clipboard_read_text_finish (clipboard, result, &error);
 
-	if (text) {
-		hex_document_set_data (self->document,
-				self->cursor_pos,
-				strlen(text),
-				0,	/* rep_len (0 to insert w/o replacing; what we want) */
-				text,
-				TRUE);
-
-		hex_widget_set_cursor (self, self->cursor_pos + strlen(text));
+	if (text)
+	{
+		paste_set_data(self, text, strlen(text));
 		g_free(text);
 	}
-	else {
+	else
+	{
 		g_critical ("Error pasting text: %s",
 				error->message);
 		g_error_free (error);
@@ -1600,22 +1662,15 @@ paste_helper (HexWidget *self, GdkClipboard *clipboard)
 	if (have_hex_paste_data)
 	{
 		char *doc_data;
-		int elems;
+		size_t start_pos, len, rep_len;
 
 		g_debug("%s: We HAVE HexPasteData.", __func__);
 
 		paste = HEX_PASTE_DATA(g_value_get_object (&value));
 		doc_data = hex_paste_data_get_doc_data (paste);
-		elems = hex_paste_data_get_elems (paste);
+		len = hex_paste_data_get_elems (paste);
 
-		hex_document_set_data (self->document,
-				self->cursor_pos,
-				elems,
-				0,	/* rep_len (0 to insert w/o replacing; what we want) */
-				doc_data,
-				TRUE);
-
-		hex_widget_set_cursor (self, self->cursor_pos + elems);
+		paste_set_data(self, doc_data, len);
 	}
 	else
 	{
@@ -1823,7 +1878,11 @@ key_press_cb (GtkEventControllerKey *controller,
 
 	show_cursor (self, FALSE);
 
-	/* Figure out if we're holding shift or not. */
+	/* don't trample over Ctrl or Alt (reserved for actions) */
+	if (state & (GDK_CONTROL_MASK | GDK_ALT_MASK)) {
+		return FALSE;
+	}
+
 	if (state & GDK_SHIFT_MASK)
 		self->selecting = TRUE;
 	else
@@ -1871,62 +1930,41 @@ key_press_cb (GtkEventControllerKey *controller,
 			break;
 
 		case GDK_KEY_Page_Up:
-			if (! (state & GDK_CONTROL_MASK))
-			{
-				hex_widget_set_cursor (self, MAX (0, self->cursor_pos - self->vis_lines*self->cpl));
-				ret = GDK_EVENT_STOP;
-			}
+			hex_widget_set_cursor (self, MAX (0, self->cursor_pos - self->vis_lines*self->cpl));
+			ret = GDK_EVENT_STOP;
 			break;
 
 		case GDK_KEY_Page_Down:
-			if (! (state & GDK_CONTROL_MASK))
-			{
-				hex_widget_set_cursor(self, MIN (payload_size,
-							self->cursor_pos + self->vis_lines*self->cpl));
-				ret = GDK_EVENT_STOP;
-			}
+			hex_widget_set_cursor(self, MIN (payload_size,
+						self->cursor_pos + self->vis_lines*self->cpl));
+			ret = GDK_EVENT_STOP;
 			break;
 
 		case GDK_KEY_Home:
-			if (state & GDK_CONTROL_MASK)
-			{
-				hex_widget_set_cursor (self, 0);
-			}
-			else
-			{
-				gint64 line_beg = self->cursor_pos; 
+		{
+			gint64 line_beg = self->cursor_pos;
 
-				while (line_beg % self->cpl != 0)
-					--line_beg;
+			while (line_beg % self->cpl != 0)
+				--line_beg;
 
-				hex_widget_set_cursor (self, line_beg);
-			}
+			hex_widget_set_cursor (self, line_beg);
 			ret = GDK_EVENT_STOP;
+		}
 			break;
 
 		case GDK_KEY_End:
-			if (state & GDK_CONTROL_MASK)
-			{
-				hex_widget_set_cursor (self, payload_size);
-			}
-			else
-			{
-				gint64 line_end = self->cursor_pos; 
+		{
+			gint64 line_end = self->cursor_pos;
 
-				while (line_end % self->cpl != self->cpl - 1)
-					++line_end;
+			while (line_end % self->cpl != self->cpl - 1)
+				++line_end;
 
-				hex_widget_set_cursor (self, MIN (line_end, payload_size));
-			}
+			hex_widget_set_cursor (self, MIN (line_end, payload_size));
 			ret = GDK_EVENT_STOP;
+		}
 			break;
 
 		default:
-			/* don't trample over Ctrl or Alt (reserved for actions) */
-			if (state & GDK_CONTROL_MASK || state & GDK_ALT_MASK) {
-				return FALSE;
-			}
-
 			if (self->active_view == VIEW_HEX)
 			{
 				switch(keyval)
@@ -2554,6 +2592,18 @@ document_changed_cb (HexDocument* doc, gpointer change_data,
 			"gtkhex.redo", hex_document_can_redo (doc));
 }
 
+static void
+file_save_started_cb (HexWidget *self, HexDocument *doc)
+{
+	set_busy_state (self, TRUE);
+}
+
+static void
+file_saved_cb (HexWidget *self, HexDocument *doc)
+{
+	set_busy_state (self, FALSE);
+}
+
 static void
 hex_widget_constructed (GObject *object)
 {
@@ -2576,6 +2626,12 @@ hex_widget_constructed (GObject *object)
 	g_signal_connect (self->document, "redo",
 			G_CALLBACK (doc_undo_redo_cb), self);
 
+	g_signal_connect_swapped (self->document, "file-save-started",
+			G_CALLBACK (file_save_started_cb), self);
+
+	g_signal_connect_swapped (self->document, "file-saved",
+			G_CALLBACK (file_saved_cb), self);
+
 	/* Chain up */
 	G_OBJECT_CLASS(hex_widget_parent_class)->constructed (object);
 }
@@ -2788,6 +2844,10 @@ hex_widget_class_init (HexWidgetClass *klass)
 			NULL,
 			geometry_action);
 
+	gtk_widget_class_install_action (widget_class, "gtkhex.move-to-buffer-ends",
+			"(bb)",
+			move_to_buffer_ends_action);
+
 	/* SHORTCUTS FOR ACTIONS (not to be confused with keybindings, which are
 	 * set up in hex_widget_init) */
 
@@ -2839,6 +2899,34 @@ hex_widget_class_init (HexWidgetClass *klass)
 			GDK_ALT_MASK,
 			"gtkhex.toggle-ascii",
 			NULL);
+
+	/* Ctrl+Home - move to beginning of buffer */
+	gtk_widget_class_add_binding_action (widget_class,
+			GDK_KEY_Home,
+			GDK_CONTROL_MASK,
+			"gtkhex.move-to-buffer-ends",
+			"(bb)", TRUE, FALSE);
+
+	/* Ctrl+End - move to end of buffer */
+	gtk_widget_class_add_binding_action (widget_class,
+			GDK_KEY_End,
+			GDK_CONTROL_MASK,
+			"gtkhex.move-to-buffer-ends",
+			"(bb)", FALSE, FALSE);
+
+	/* Shift+Ctrl+Home - move to beginning of buffer with selection */
+	gtk_widget_class_add_binding_action (widget_class,
+			GDK_KEY_Home,
+			GDK_SHIFT_MASK | GDK_CONTROL_MASK,
+			"gtkhex.move-to-buffer-ends",
+			"(bb)", TRUE, TRUE);
+
+	/* Shift+Ctrl+End - move to end of buffer with selection */
+	gtk_widget_class_add_binding_action (widget_class,
+			GDK_KEY_End,
+			GDK_SHIFT_MASK | GDK_CONTROL_MASK,
+			"gtkhex.move-to-buffer-ends",
+			"(bb)", FALSE, TRUE);
 }
 
 static void
@@ -3197,6 +3285,15 @@ hex_widget_cut_to_clipboard (HexWidget *self)
  * hex_widget_paste_from_clipboard:
  *
  * Paste clipboard data to widget at position of cursor.
+ *
+ * Since 4.6, the behaviour of this method has changed. With 4.4 and earlier,
+ * paste operations always inserted data into the payload, even if insert mode
+ * was disabled.
+ *
+ * Commencing in 4.6, if insert mode is not enabled, data will be overwritten
+ * by default with a paste operation, and possibly truncated in the event the
+ * payload is not large enough to absorb the paste data. This is to avoid
+ * increasing the payload size of a hex document when insert mode is disabled.
  */
 void
 hex_widget_paste_from_clipboard (HexWidget *self)
@@ -3806,3 +3903,5 @@ hex_widget_get_group_type (HexWidget *self)
 
 	return self->group_type;
 }
+
+G_GNUC_END_IGNORE_DEPRECATIONS
